//Title: MediaPlayerNode Class
//Platform: BeOS 5
//Version: 2002.03.19
//Description:
//	A class that creates a BBufferProducer Media Node, and plays media from a BMediaFile.
//
	
//Copyright (C) 2001, 2002 Kevin H. Patterson
//Available for use under the terms of the BSD license agreement:
//See the associated file "License.txt" for details.

#ifndef _MediaPlayerNode_h
#define _MediaPlayerNode_h
	
	#include "globals.h"

	#include <Application.h>
	#include <MediaFile.h>
	#include <Message.h>
	#include <MediaNode.h>
	#include <MediaEventLooper.h>
	#include <TimeSource.h>
	#include <Buffer.h>
	#include <BufferProducer.h>
	#include <BufferGroup.h>
	#include <Locker.h>
	
	#include "OutputStream.h"
	
	class MediaPlayerNode
	:
		public BBufferProducer,
		public BMediaEventLooper
	{
	
	public:
		MediaPlayerNode( const char *uMediaFileName );
		~MediaPlayerNode();
		status_t InitCheck() const;

	private:
		status_t iInitCheck;
		BMediaFile * iMediaFile;
		int32	iMediaStreamCount;
		OutputStream * iMediaStream[64];
		volatile bool iRunning;
		BLocker * iNodeLock;
		
	// BMediaNode
	public:
		port_id ControlPort() const;
		BMediaAddOn * AddOn( int32 * outInternalID ) const;
		status_t HandleMessage(
			int32 message,
			const void * data,
			size_t size );
		
	protected:
		void Preroll();
		void SetTimeSource( BTimeSource * time_source );
		status_t RequestCompleted( const media_request_info & info );
		
	// BBufferProducer
	protected:
		status_t FormatSuggestionRequested(
			media_type type,
			int32 quality,
			media_format * format );
		status_t FormatProposal(
			const media_source & output,
			media_format * format );
		status_t FormatChangeRequested(
			const media_source & source,
			const media_destination & destination,
			media_format * format,
			int32 * _deprecated_ );
		status_t GetNextOutput(
			int32 * cookie,
			media_output * outOutput );
		status_t DisposeOutputCookie( int32 cookie );
		status_t SetBufferGroup(
			const media_source & source,
			BBufferGroup * group );
		status_t VideoClippingChanged(
			const media_source & source,
			int16 num_shorts,
			int16 * clip_data,
			const media_video_display_info & display,
			int32 * _deprecated_ );
		status_t GetLatency( bigtime_t * latency );
		status_t PrepareToConnect(
			const media_source & what,
			const media_destination & where,
			media_format * format,
			media_source * outSource,
			char * outName );
		void Connect(
			status_t error,
			const media_source & source,
			const media_destination & destination,
			const media_format & format,
			char * outName );
		void Disconnect(
			const media_source & what,
			const media_destination & where );
		void LateNoticeReceived(
			const media_source & what,
			bigtime_t how_much,
			bigtime_t performance_time );
		void EnableOutput(
			const media_source & what,
			bool enabled,
			int32 * _deprecated_ );
		status_t SetPlayRate( int32 numer, int32 denom );
		void AdditionalBufferRequested(
			const media_source & source,
			media_buffer_id prev_buffer,
			bigtime_t prev_time,
			const media_seek_tag * prev_tag );
		void LatencyChanged(
			const media_source & source,
			const media_destination & destination,
			bigtime_t new_latency,
			uint32 flags);
		
	// BMediaEventLooper
	protected:
		void NodeRegistered();
		void Start( bigtime_t performance_time );
		void Stop( bigtime_t performance_time, bool immediate );
		void Seek( bigtime_t media_time, bigtime_t performance_time );
		void TimeWarp( bigtime_t at_real_time, bigtime_t to_performance_time );
		status_t AddTimer( bigtime_t at_performance_time, int32 cookie );
		void SetRunMode( run_mode mode );
		void HandleEvent(
			const media_timed_event *event,
			bigtime_t lateness,
			bool realTimeEvent = false );
		void CleanUpEvent( const media_timed_event *event );
		bigtime_t OfflineTime();
		void ControlLoop();
		status_t DeleteHook( BMediaNode * node );

	// Implementation
	public:
		status_t iSendBuffer( BBuffer * buffer, const media_destination & destination );
		void iOutputLatencyChanged( bigtime_t new_latency );
		
	private:
		void HandleStart( bigtime_t performance_time );
		void HandleStop();
		void HandleStop( bigtime_t performance_time );
		void HandleTimeWarp( bigtime_t performance_time );
		void HandleSeek( bigtime_t media_time, bigtime_t performance_time );
		bigtime_t RecomputeLatency();
		
	public:
		bigtime_t Duration() const;
		bigtime_t Position() const;
	};
	
#endif
