/* BT868_9.cpp
** written by Rudolf Cornelissen
** Email: rag.cornelissen@inter.nl.net
*/

#include <stdio.h>
#include <malloc.h>
#include <module.h>
#include <Drivers.h>

#include "../riva_iic/riva_iic.h"
#include "BT868_9.h"

	const unsigned char PRADR = 0x88,
						SCADR = 0x8a,
						WR = 0x00,
						RD = 0x01,
						FAKE = 0xee;

byte 	Bus,	//IIC busnr on which BT resides
		Adres;	//IIC adres on which BT resides
int 	bus0sane = 0,
		bus1sane = 0;

int ReadType (int fd, int *Version, int *Type, riva_data_t *ddata)
{
	byte buffer[3];
	int rd, err;

	ddata->iic.buffer = &buffer;

	//Make sure a CX (Conexant) chip (if this turns out to be there) is set to
	//BT-compatibility mode! (This command will do nothing on a BT chip...)
	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0x6c;		//select CX reg. for BT-compatible readback, video still off
	buffer[2] = 0x02;		//set it up.
	
	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 3); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
		goto ChkForRegroup;

	//Do actual readtype command:
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit 1 byte
	buffer[0] = Adres + RD;	//issue IIC read command
	write(fd, ddata, 1); 
	//receive 1 byte
	//ACK level to TX after last byte to RX should be 1 (= NACK) (see IIC spec)
	//While the BT's don't care, CX chips will block the SDA line if an ACK gets sent!
	buffer[0] = 1;	
	rd = read(fd, ddata, 1);
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (!rd==1 || err) //(not one byte RX'd) or (IIC comms error)?
		goto ChkForRegroup;

	*Type = (buffer[0] & 0xe0) >> 5;
	*Version = buffer[0] & 0x1f;
	return (0);

ChkForRegroup:
	//NACK: occurs on some ASUS V7700 GeForce cards!
	//(apparantly the video-in chip or another chip resides at 'BT' adresses there..)
	if (ddata->iic.status == 3)
	{
		printf ("\nChip does not respond as expected, ignoring this chip.\n");
		return(2);
	}
	else
	{
		printf ("\nIIC communication error %d encountered, exiting...\n", ddata->iic.status);
		close (fd);
		return(1);
	}
}

int FindSaneBus (int fd, int bus0wired, int bus1wired,
										int *bus0valid, int *bus1valid, riva_data_t *ddata)
{
	int err;//, Bus = 0;
	byte buffer[1];
	ddata->iic.buffer = &buffer;

	if (bus0wired)
	{
		printf("IIC bus 0 sanity check: ");
		ddata->iic.busnr = 0;	//IIC bus 0
		ioctl(fd, rivaStartIIC, ddata);//iicstart); 
		buffer[0] = FAKE+WR;	//primary adress
		write(fd, ddata, 1); 
		err = ioctl(fd, rivaStopIIC, ddata);//iicstop); 
		if (err)	//if there is no error, the bus must be insane! (Bus should always be sane..)
		{
			printf("OK. (bus status = %d)\n",ddata->iic.status);
			bus0sane = 1;
			*bus0valid = 1;
		}
		else
		{
			printf("fail\n");
		}
	}
	
	if (bus1wired)
	{
		printf("IIC bus 1 sanity check: ");
		ddata->iic.busnr = 1;
		ioctl(fd, rivaStartIIC, ddata);//iicstart); 
		buffer[0] = FAKE+WR;//primary adress
		write(fd, ddata, 1); 
		err = ioctl(fd, rivaStopIIC, ddata);//iicstop); 
		if (err)
		{
			printf("OK. (bus status = %d)\n",ddata->iic.status);
			bus1sane = 1;
			*bus1valid = 1;
		}
		else
		{
			printf("fail\n");
		}
	}
	if (!bus0sane && !bus1sane) return (0);
	else return (1);
}

int FindBT (int fd, riva_data_t *ddata)
{
	int btfound = 0;

	printf("\nChecking sane IIC bus(ses) for first possible TV-out chip: ");
	if (bus0sane)
	{
		//try primary adress on bus 0:
		ddata->iic.busnr = 0; //IIC bus to use for this attempt
		if (!ChkForBT (fd, PRADR, ddata))
		{
			btfound = 1;
			Adres = PRADR;
			Bus = 0; //IIC bus on which the TVout chip was found
		}
		else
		{
			//try secondary adress on bus 0:
			if (!ChkForBT (fd, SCADR, ddata))
			{
				btfound = 1;
				Adres = SCADR;
				Bus = 0;
			}
		}
	}
	if (bus1sane && !btfound)
	{
		//try primary adress on bus 1:
		ddata->iic.busnr = 1;
		if (!ChkForBT (fd, PRADR, ddata))
		{
			btfound = 1;
			Adres = PRADR;
			Bus = 1;
		}
		else
		{
			//try secondary adress on bus 1:
			if (!ChkForBT (fd, SCADR, ddata))
			{
				btfound = 1;
				Adres = SCADR;
				Bus = 1;
			}
		}
	}
	if (btfound) printf("\nFound one on bus %d, adress $%x;\n",Bus,Adres);
	return (btfound);
}

int ChkForBT (int fd, int Adres, riva_data_t *ddata)
{
	byte buffer[3];
	ddata->iic.buffer = &buffer;

	ioctl(fd, rivaStartIIC, ddata);//iicstart); 
	buffer[0] = Adres + WR;//primary adress
	buffer[1] = 0xc4; //set ESTATUS at b'00'; and enable bt chip-outputs
//WARNING: 	If bit0 = 0 is issued below (EN_OUT = disabled), the BT will lock SDA
//			after writing adress $A0 (setting EN_XCLK)!!!
//			Until a reboot the corresponding IIC bus will be inacessable then!!!

	//fixme: if testimage 'was' active txbuffer[3] should become 0x05...
	//(currently this cannot be detected in a 'foolproof' way so don't touch...)
	buffer[2] = 0x01; //(ESTATUS b'0x' means: RX ID and VERSION info later..)
	write(fd, ddata, 3); 
	return ioctl(fd, rivaStopIIC, ddata);//iicstop); 
}

int DoTestImage (int fd, riva_data_t *ddata)
{
	int err;

	byte buffer[3];
	ddata->iic.buffer = &buffer;

	//activate actual testimage (colorbars)...
	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0xc4;		//select bt register for enabling colorbars and outputs
	buffer[2] = 0x05;		//issue the actual command

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 3); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return (err);
}//end DoTestImage.

int SetupBTforTVout (int fd, int OutMethod, int monstat, int ffilter, riva_data_t *ddata)
{
	int err;
	byte buffer[7];
	ddata->iic.buffer = &buffer;

	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0xc6;		//select first TV config register
	buffer[2] = 0x98;		//input is 24bit mpx'd RGB, BLANK = out, sync = act. hi
	buffer[3] = 0x98;		//disable all filters, exept flicker filter.
	if (!ffilter)
	{
		buffer[3] = 0xc0;	//disable flicker filter.
	}

	buffer[4] = 0xc0;		//(disable filters)
	buffer[5] = 0xc0;		//(disable filters)
	switch (OutMethod)
	//Description of ELSA Erazor III hardware layout:
	//(This is the default (recommended) layout by NVIDIA)
	//DAC A = CVBS
	//DAC B = C (chrominance)
	//DAC C = Y (luminance)/
		
	//Description of Diamond VIPER550:
	//DAC A = Not connected
	//DAC B = C (chrominance)
	//DAC C = Y (luminance)
	//To be able to connect to CVBS TV's a special cable is supplied:
	//This cable connects the Y (DAC C) output to the TV CVBS input.
	{
		case 1: buffer[6] = 0x18;	// Y/C and CVBS out if all ports implemented
									// in hardware, else only Y/C or CVBS out.
				break;
		case 2: buffer[6] = 0x00;	// put CVBS on all outputs. Used for cards
				break;				// with only Y/C out and 'translation cable'.
		default:switch (monstat)	// only 'autodetect' remains...
				{
					case 1: buffer[6] = 0x00;	//only Y connected: must be CVBS!
							break;
					case 2: buffer[6] = 0x00;	//only C connected: must be CVBS!
							break;				//(though cable is wired wrong...)
					case 5: buffer[6] = 0x00;	//CVBS and only Y connected: 2x CVBS!
							break;			   	//(officially not supported...)
					case 6: buffer[6] = 0x00;	//CVBS and only C connected: 2x CVBS!
							break;			   	//(officially not supported...)
					default:buffer[6] = 0x18;	//nothing, or
											 	//Y/C only, or
											 	//CVBS only (but on CVBS output), or
											 	//Y/C and CVBS connected:
											 	//So activate recommended signals.
				}		
	}	

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 7); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return (err);
}//end SetupBTforTVout.

int SetBT_Hphase (int fd, int mode, riva_data_t *ddata)
{
	int err;
	
	byte Hoffset, buffer[7];
	ddata->iic.buffer = &buffer;

	//values below are all tested on TNT1 and TNT2.
	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0x6c;		//select first TV output timing register
	buffer[2] = 0x06;		//turn on active video on CX chips!
	//(fail save values...)
	buffer[3] = 0x00;		//set default horizontal sync offset
	buffer[4] = 0x02;		//set default horizontal sync width
	buffer[5] = 0x00;		//set default vertical sync offset

	//do specific timing setup for all chips and modes:
	Hoffset = 8;					//TNT2 and TNT2 M64...(8 pixels delayed hpos,
									//so picture more to the right)
	if (ddata->type == 1) Hoffset = 0;	//TNT1 or GeForce2... (std hpos)
	switch (mode)
		{
		case 1: 
		case 2: buffer[3] = (0x25 + Hoffset);	//set horizontal sync offset
				break;
		case 3: buffer[3] = (0xe1 + Hoffset);	//set horizontal sync offset
				buffer[4] = 0xc2;
				//WARNING: The VSync offset register does not exist in CX chips!
				//This mode has still to be checked for stable and centered output
				//on such a chip...
				buffer[5] = 0x40;				//set VSync offset (on BT's only)
				break;
		case 5: buffer[3] = (0xa8 + Hoffset);
				break;
		case 4: 
		case 6: buffer[3] = (0x2c + Hoffset);
				break;
		case 9: buffer[3] = (0xd0 + Hoffset);	//set horizontal sync offset 
				buffer[4] = 0xff;				//hsync width = max:
				break;							//to prevent vertical image 'shivering'.
		case 10:buffer[3] = (0xd4 + Hoffset);
				buffer[4] = 0xff;
				break;
		case 11:buffer[3] = (0xc8 + Hoffset);
				buffer[4] = 0xff;
				break;
		case 12:buffer[3] = (0xc4 + Hoffset);
				buffer[4] = 0xff;
				break;
		default:;//nothing to be done here...
		}		

	buffer[6] = 0x01;		//set default vertical sync width

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 7); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return(err);
}//end SetBT_Hphase.

int InitBT_PAL640 (int fd, riva_data_t *ddata)
{
	int err;
	
	byte buffer[35];
	ddata->iic.buffer = &buffer;

	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0x76;			//select first bt register to write to
	buffer[2] = 0x60;
	buffer[3] = 0x80;
	buffer[4] = 0x8a;
	buffer[5] = 0xa6;
	buffer[6] = 0x68;
	buffer[7] = 0xc1;
	buffer[8] = 0x2e;
	buffer[9] = 0xf2;
	buffer[10] = 0x27;
	buffer[11] = 0x00;
	buffer[12] = 0xb0;
	buffer[13] = 0x0a;
	buffer[14] = 0x0b;
	buffer[15] = 0x71;
	buffer[16] = 0x5a;
	buffer[17] = 0xe0;
	buffer[18] = 0x36;
	buffer[19] = 0x00;
	buffer[20] = 0x50;
	buffer[21] = 0x72;
	buffer[22] = 0x1c;
	buffer[23] = 0x8d;		//chip-pin CLKI is pixel clock (only non-default here!)
	buffer[24] = 0x24;
	buffer[25] = 0xf0;
	buffer[26] = 0x58;
	buffer[27] = 0x81;
	buffer[28] = 0x49;
	buffer[29] = 0x8c;
	buffer[30] = 0x0c;
	buffer[31] = 0x8c;
	buffer[32] = 0x79;
	buffer[33] = 0x26;
	buffer[34] = 0x00;

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 35); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return(err);
}//end InitBT_PAL640.

int InitBT_PAL800 (int fd, riva_data_t *ddata)
{
	int err;
	
	byte buffer[35];
	ddata->iic.buffer = &buffer;

	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0x76;			//select first bt register to write to
	buffer[2] = 0x00;
	buffer[3] = 0x20;
	buffer[4] = 0xaa;
	buffer[5] = 0xca;
	buffer[6] = 0x9a;
	buffer[7] = 0x0d;
	buffer[8] = 0x29;
	buffer[9] = 0xfc;
	buffer[10] = 0x39;
	buffer[11] = 0x00;
	buffer[12] = 0xc0;
	buffer[13] = 0x8c;
	buffer[14] = 0x03;
	buffer[15] = 0xee;
	buffer[16] = 0x5f;
	buffer[17] = 0x58;
	buffer[18] = 0x3a;
	buffer[19] = 0x66;
	buffer[20] = 0x96;
	buffer[21] = 0x00;
	buffer[22] = 0x00;
	buffer[23] = 0x90;		//chip-pin CLKI is pixel clock (only non-default here!)
	buffer[24] = 0x24;
	buffer[25] = 0xf0;
	buffer[26] = 0x57;
	buffer[27] = 0x80;
	buffer[28] = 0x48;
	buffer[29] = 0x8c;
	buffer[30] = 0x18;
	buffer[31] = 0x28;
	buffer[32] = 0x87;
	buffer[33] = 0x1f;
	buffer[34] = 0x00;

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 35); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return(err);
}//end InitBT_PAL800.

int InitBT_NTSC640 (int fd, riva_data_t *ddata)
{
	int err;
	
	byte buffer[35];
	ddata->iic.buffer = &buffer;

	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0x76;			//select first bt register to write to
	buffer[2] = 0x00;
	buffer[3] = 0x80;
	buffer[4] = 0x84;
	buffer[5] = 0x96;
	buffer[6] = 0x60;
	buffer[7] = 0x7d;
	buffer[8] = 0x22;
	buffer[9] = 0xd4;
	buffer[10] = 0x27;
	buffer[11] = 0x00;
	buffer[12] = 0x10;
	buffer[13] = 0x7e;
	buffer[14] = 0x03;
	buffer[15] = 0x58;
	buffer[16] = 0x4b;
	buffer[17] = 0xe0;
	buffer[18] = 0x36;
	buffer[19] = 0x92;
	buffer[20] = 0x54;
	buffer[21] = 0x0e;
	buffer[22] = 0x88;
	buffer[23] = 0x8c;		//chip-pin CLKI is pixel clock (only non-default here!)
	buffer[24] = 0x0a;
	buffer[25] = 0xe5;
	buffer[26] = 0x76;
	buffer[27] = 0x79;
	buffer[28] = 0x44;
	buffer[29] = 0x85;
	buffer[30] = 0x00;
	buffer[31] = 0x00;
	buffer[32] = 0x80;
	buffer[33] = 0x20;
	buffer[34] = 0x00;

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 35); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return(err);
}//end InitBT_NTSC640.

int InitBT_NTSC800 (int fd, riva_data_t *ddata)
{
	int err;
	
	byte buffer[35];
	ddata->iic.buffer = &buffer;

	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0x76;			//select first bt register to write to
	buffer[2] = 0xa0;
	buffer[3] = 0x20;
	buffer[4] = 0xb6;
	buffer[5] = 0xce;
	buffer[6] = 0x84;
	buffer[7] = 0x55;
	buffer[8] = 0x20;
	buffer[9] = 0xd8;
	buffer[10] = 0x39;
	buffer[11] = 0x00;
	buffer[12] = 0x70;
	buffer[13] = 0x42;
	buffer[14] = 0x03;
	buffer[15] = 0xdf;
	buffer[16] = 0x56;
	buffer[17] = 0x58;
	buffer[18] = 0x3a;
	buffer[19] = 0xcd;
	buffer[20] = 0x9c;
	buffer[21] = 0x14;
	buffer[22] = 0x3b;
	buffer[23] = 0x91;		//chip-pin CLKI is pixel clock (only non-default here!)
	buffer[24] = 0x0a;
	buffer[25] = 0xe5;
	buffer[26] = 0x74;
	buffer[27] = 0x77;
	buffer[28] = 0x43;
	buffer[29] = 0x85;
	buffer[30] = 0xba;
	buffer[31] = 0xe8;
	buffer[32] = 0xa2;
	buffer[33] = 0x17;
	buffer[34] = 0x00;

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 35); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return(err);
}//end InitBT_NTSC800.

int InitBT_PAL720 (int fd, riva_data_t *ddata)
{
	int err;
	
	byte buffer[35];
	ddata->iic.buffer = &buffer;

	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0x76;			//select first bt register to write to
	buffer[2] = 0xf0;
	buffer[3] = 0xd0;
	buffer[4] = 0x82;
	buffer[5] = 0x9c;
	buffer[6] = 0x5a;
	buffer[7] = 0x31;
	buffer[8] = 0x16;
	buffer[9] = 0x22;
	buffer[10] = 0xa6;
	buffer[11] = 0x00;
	buffer[12] = 0x78;
	buffer[13] = 0x93;
	buffer[14] = 0x03;
	buffer[15] = 0x71;
	buffer[16] = 0x2a;
	buffer[17] = 0x40;
	buffer[18] = 0x0a;
	buffer[19] = 0x00;
	buffer[20] = 0x50;
	buffer[21] = 0x55;
	buffer[22] = 0x55;
	buffer[23] = 0x8c;		//chip-pin CLKI is pixel clock (only non-default here!)
	buffer[24] = 0x24;
	buffer[25] = 0xf0;
	buffer[26] = 0x59;
	buffer[27] = 0x82;
	buffer[28] = 0x49;
	buffer[29] = 0x8c;
	buffer[30] = 0x8e;
	buffer[31] = 0xb0;
	buffer[32] = 0xe6;
	buffer[33] = 0x28;
	buffer[34] = 0x00;

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 35); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return(err);
}//end InitBT_PAL720.

int InitBT_NTSC720 (int fd, riva_data_t *ddata)
{
	int err;
	
	byte buffer[35];
	ddata->iic.buffer = &buffer;

	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0x76;		//select first bt register to write to.
	buffer[2] = 0xf0;		//lsb h_clk_o: overscan comp = 0, so h_clk_o = 2 * h_clk_i (VSR=2 = scaling=1)
	buffer[3] = 0xd0;		//lsb h_active: h_active = 720 pixels wide port
	buffer[4] = 0x83;	//scope: OK hsync_width: (hsync_width / h_clk_o) * 63.55556uS = 4.70uS for NTSC
	buffer[5] = 0x98;	//scope: OK	hburst_begin: (hburst_begin / h_clk_o) * 63.55556uS = 5.3uS for NTSC
	buffer[6] = 0x5e;	//scope: OK hburst_end: ((hburst_end + 128) / h_clk_o) * 63.55556uS = 7.94uS for NTSC

	//How to find the correct values for h_blank_o and v_blank_o:
	// 1. Calculate h_blank_o according to initial setting guideline mentioned below;
	// 2. Set v_blank_o in the neighbourhood of $18, so that TV picture does not have ghosts on right side in it while
	//    horizontal position is about OK;
	// 3. Then tune h_blank_o for centered output on scope (look at front porch and back porch);
	// 4. Now shift the TV output using Hsync_offset for centered output while looking at TV (in method 'SetBT_Hphase' above);
	// 5. If no vertical shivering occurs when image is centered, you're done. Else:
	// 6. Modify the RIVA (BeScreen) h_sync_start setting somewhat to get stable centered picture possible on TV AND!:
	// 7. Make sure you update the Chrontel horizontal Phase setting also then!

	buffer[7] = 0x28;	//scope: tuned. lsb h_blank_o: h_blank_o = horizontal viewport location on TV
						//(guideline for initial setting: (h_blank_o / h_clk_0) * 63.55556uS = 9.5uS for NTSC)
	buffer[8] = 0x18;	//try-out; scope: checked against other modes, looks OK.	v_blank_o: 1e active line ('pixel')

	buffer[9] = 0xf2;		//v_active_o: = (active output lines + 2) / field (on TV)
	buffer[10] = 0x26;		//lsn = msn h_clk_o;
							//b4-5 = msbits h_active;
							//b7 = b8 v_avtive_o.
	buffer[11] = 0x00;		//h_fract is always 0.
	buffer[12] = 0x78;		//lsb h_clk_i: h_clk_i is horizontal total = 888.
	buffer[13] = 0x90; 	//try-out; lsb h_blank_i: #clks between start sync and new line 1st pixel; copy to VGA delta-sync!
	buffer[14] = 0x03;		//b2-0 = msn h_clk_i;
						//try-out: b3 = msn h_blank_i;
							//b4 = vblankdly is always 0.
	buffer[15] = 0x0d;		//lsb v_lines_i: v_lines_i = 525
	buffer[16] = 0x1a; 	//try-out;	v_blank_i: #input lines between start sync and new line (pixel); copy to VGA delta-sync!
						//Make sure though that this value for the BT is *even*, or image will shiver a *lot* horizontally on TV.
	buffer[17] = 0xe0;		//lsb v_active_i: v_active_i = 480
	buffer[18] = 0x36;		//b1-0 = msn v_lines_i;
							//b3-2 = msn v_active_i;
							//b5-4 = ylpf = 3;
							//b7-6 = clpf = 0.		
	buffer[19] = 0x00;		//lsb v_scale: v_scale = off = $1000
	buffer[20] = 0x50;		//b5-0 = msn v_scale;
						//scope: tuned. b7-6 = msn h_blank_o.
							//(((PLL_INT + (PLL_FRACT/65536)) / 6) * 13500000) = PIXEL_CLK = (hor.tot. * v_lines_i * 60Hz)
	buffer[21] = 0x98;		//lsb PLL fract: PLL fract = 0x6e98
	buffer[22] = 0x6e;		//msb PLL fract
	buffer[23] = 0x8c;		//b5-0 = PLL int: PLL int = 0x0c;
							//b6 = by_pll: by_pll = 0;
							//b7 = EN_XCLK: chip-pin CLKI is pixel clock.
	buffer[24] = 0x0a;		//b0 = ni_out is always 0;
							//b1 = setup = 1 for NTSC;
							//b2 = 625line = 0 for NTSC;
							//b3 = vsync_dur = 1 for NTSC;
							//b4 = dic_screset is always 0;
							//b5 = pal_md = 0 for NTSC;
							//b6 = eclip is always 0;
							//b7 = reserved (en_scart) is always 0.
	buffer[25] = 0xe5;		//sync_amp $e5 for NTSC
	buffer[26] = 0x75;		//bst_amp $74-$76 for NTSC
	buffer[27] = 0x78;		//mcr: r-y $77-$79 for NTSC
	buffer[28] = 0x44;		//mcb: b-y $43-$44 for NTSC
	buffer[29] = 0x85;		//my: y $85 for NTSC
	buffer[30] = 0x3c;		//lsb msc: msc b31-0: NTSC formula: ((3579545 / pixelclk) * 2^32) = MSC
	buffer[31] = 0x91;		//msc = $20c2913c
	buffer[32] = 0xc2;
	buffer[33] = 0x20;		//msb msc.
	buffer[34] = 0x00;		//phase_off always $00

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 35); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return(err);
}//end InitBT_NTSC720.

int InitBT_PAL800_OS (int fd, riva_data_t *ddata)
{
	int err;
	
	byte buffer[35];
	ddata->iic.buffer = &buffer;

	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0x76;		//select first bt register to write to.
	buffer[2] = 0x60;		//lsb h_clk_o: overscan comp = 0, so h_clk_o = 2 * h_clk_i (VSR=2 = scaling=1)
	buffer[3] = 0x20;		//lsb h_active: h_active = 800 pixels wide port
	buffer[4] = 0x8b;	//scope: OK hsync_width: (hsync_width / h_clk_o) * 64.0uS = 4.70uS for PAL
	buffer[5] = 0xa5;	//scope: OK	hburst_begin: (hburst_begin / h_clk_o) * 64.0uS = 5.6uS for PAL
	buffer[6] = 0x6b;	//scope: OK hburst_end: ((hburst_end + 128) / h_clk_o) * 64.0uS = 7.97uS for PAL

	//How to find the correct values for h_blank_o and v_blank_o:
	// 1. Calculate h_blank_o according to initial setting guideline mentioned below;
	// 2. Set v_blank_o in the neighbourhood of $18, so that TV picture does not have ghosts on right side in it while
	//    horizontal position is about OK;
	// 3. Then tune h_blank_o for centered output on scope (look at front porch and back porch);
	// 4. Now shift the TV output using Hsync_offset for centered output while looking at TV (in method 'SetBT_Hphase' above);
	// 5. If no vertical shivering occurs when image is centered, you're done. Else:
	// 6. Modify the RIVA (BeScreen) h_sync_start setting somewhat to get stable centered picture possible on TV AND!:
	// 7. Make sure you update the Chrontel horizontal Phase setting also then!

	buffer[7] = 0xc0;	//scope: tuned. lsb h_blank_o: h_blank_o = horizontal viewport location on TV
						//(guideline for initial setting: (h_blank_o / h_clk_0) * 64.0uS = 10.0uS for PAL)
	buffer[8] = 0x18;	//try-out; scope: checked against other modes, looks OK.	v_blank_o: 1e active line ('pixel')

	buffer[9] = 0x2e;		//v_active_o: = (active output lines + 2) / field (on TV)
	buffer[10] = 0xb7;		//lsn = msn h_clk_o;
							//b4-5 = msbits h_active;
							//b7 = b8 v_avtive_o.
	buffer[11] = 0x00;		//h_fract is always 0.
	buffer[12] = 0xb0;		//lsb h_clk_i: h_clk_i is horizontal total = 944.
	buffer[13] = 0x14; 	//try-out; lsb h_blank_i: #clks between start sync and new line 1st pixel; copy to VGA delta-sync!
	buffer[14] = 0x03;		//b2-0 = msn h_clk_i;
					 	//try-out: b3 = msn h_blank_i;
							//b4 = vblankdly is always 0.
	buffer[15] = 0x71;		//lsb v_lines_i: v_lines_i = 625
	buffer[16] = 0x2a; 	//try-out;	v_blank_i: #input lines between start sync and new line (pixel); copy to VGA delta-sync!
					 	//Make sure though that this value for the BT is *even*, or image will shiver a *lot* horizontally on TV.
	buffer[17] = 0x58;		//lsb v_active_i: v_active_i = 600
	buffer[18] = 0x3a;		//b1-0 = msn v_lines_i;
							//b3-2 = msn v_active_i;
							//b5-4 = ylpf = 3;
							//b7-6 = clpf = 0.		
	buffer[19] = 0x00;		//lsb v_scale: v_scale = off = $1000
	buffer[20] = 0x10;		//b5-0 = msn v_scale;
						//scope: tuned. b7-6 = msn h_blank_o.
							//(((PLL_INT + (PLL_FRACT/65536)) / 6) * 13500000) = PIXEL_CLK = (hor.tot. * v_lines_i * 50Hz)
	buffer[21] = 0x72;		//lsb PLL fract: PLL fract = 0x1c72
	buffer[22] = 0x1c;		//msb PLL fract
	buffer[23] = 0x8d;		//b5-0 = PLL int: PLL int = 0x0d;
							//b6 = by_pll: by_pll = 0;
							//b7 = EN_XCLK: chip-pin CLKI is pixel clock.
	buffer[24] = 0x24;		//b0 = ni_out is always 0;
							//b1 = setup = 0 for PAL;
							//b2 = 625line = 1 for PAL;
							//b3 = vsync_dur = 0 for PAL;
							//b4 = dic_screset is always 0;
							//b5 = pal_md = 1 for PAL;
							//b6 = eclip is always 0;
							//b7 = reserved (en_scart) is always 0.
	buffer[25] = 0xf0;		//sync_amp $f0 for PAL
	buffer[26] = 0x57;		//bst_amp $57-$58 for PAL
	buffer[27] = 0x80;		//mcr: r-y $80-$81 for PAL
	buffer[28] = 0x48;		//mcb: b-y $48-$49 for PAL
	buffer[29] = 0x8c;		//my: y $8c for PAL
	buffer[30] = 0x31;		//lsb msc: msc b31-0: PAL formula: ((4433619 / pixelclk) * 2^32) = MSC
	buffer[31] = 0x8c;		//msc = $26798c31
	buffer[32] = 0x79;
	buffer[33] = 0x26;		//msb msc.
	buffer[34] = 0x00;		//phase_off always $00

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 35); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return(err);
}//end InitBT_PAL800_OS.

int InitBT_NTSC640_OS (int fd, riva_data_t *ddata)
{
	int err;
	
	byte buffer[35];
	ddata->iic.buffer = &buffer;

	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0x76;		//select first bt register to write to.
	buffer[2] = 0x20;		//lsb h_clk_o: overscan comp = 0, so h_clk_o = 2 * h_clk_i (VSR=2 = scaling=1)
	buffer[3] = 0x80;		//lsb h_active: h_active = 640 pixels wide port
	buffer[4] = 0x74;	//scope: OK hsync_width: (hsync_width / h_clk_o) * 63.55556uS = 4.70uS for NTSC
	buffer[5] = 0x83;	//scope: OK	hburst_begin: (hburst_begin / h_clk_o) * 63.55556uS = 5.3uS for NTSC
	buffer[6] = 0x44;	//scope: OK hburst_end: ((hburst_end + 128) / h_clk_o) * 63.55556uS = 7.94uS for NTSC

	//How to find the correct values for h_blank_o and v_blank_o:
	// 1. Calculate h_blank_o according to initial setting guideline mentioned below;
	// 2. Set v_blank_o in the neighbourhood of $18, so that TV picture does not have ghosts on right side in it while
	//    horizontal position is about OK;
	// 3. Then tune h_blank_o for centered output on scope (look at front porch and back porch);
	// 4. Now shift the TV output using Hsync_offset for centered output while looking at TV (in method 'SetBT_Hphase' above);
	// 5. If no vertical shivering occurs when image is centered, you're done. Else:
	// 6. Modify the RIVA (BeScreen) h_sync_start setting somewhat to get stable centered picture possible on TV AND!:
	// 7. Make sure you update the Chrontel horizontal Phase setting also then!

	buffer[7] = 0xf7;	//scope: tuned. lsb h_blank_o: h_blank_o = horizontal viewport location on TV:
						//(guideline for initial setting: (h_blank_o / h_clk_0) * 63.55556uS = 9.5uS for NTSC)
	buffer[8] = 0x1c;	//try-out; scope: checked against other modes, looks OK.	v_blank_o: 1e active line ('pixel')

	buffer[9] = 0xf2;		//v_active_o: = (active output lines + 2) / field (on TV)
	buffer[10] = 0x26;		//lsn = msn h_clk_o;
							//b4-5 = msbits h_active;
							//b7 = b8 v_avtive_o.
	buffer[11] = 0x00;		//h_fract is always 0.
	buffer[12] = 0x10;		//lsb h_clk_i: h_clk_i is horizontal total = 784.
	buffer[13] = 0x14;	//try-out; lsb h_blank_i: #clks between start sync and new line 1st pixel; copy to VGA delta-sync!
	buffer[14] = 0x03;		//b2-0 = msn h_clk_i;
						//try-out: b3 = msn h_blank_i;
							//b4 = vblankdly is always 0.
	buffer[15] = 0x0d;		//lsb v_lines_i: v_lines_i = 525
	buffer[16] = 0x18;	//try-out;	v_blank_i: #input lines between start sync and new line (pixel); copy to VGA delta-sync!
					 	//Make sure though that this value for the BT is *even*, or image will shiver a *lot* horizontally on TV.
	buffer[17] = 0xe0;		//lsb v_active_i: v_active_i = 480
	buffer[18] = 0x36;		//b1-0 = msn v_lines_i;
							//b3-2 = msn v_active_i;
							//b5-4 = ylpf = 3;
							//b7-6 = clpf = 0.		
	buffer[19] = 0x00;		//lsb v_scale: v_scale = off = $1000
	buffer[20] = 0x10;		//b5-0 = msn v_scale;
						//scope: tuned. b7-6 = msn h_blank_o.
							//(((PLL_INT + (PLL_FRACT/65536)) / 6) * 13500000) = PIXEL_CLK = (hor.tot. * v_lines_i * 60Hz)
	buffer[21] = 0xdb;		//lsb PLL fract: PLL fract = 0xf9db
	buffer[22] = 0xf9;		//msb PLL fract
	buffer[23] = 0x8a;		//b5-0 = PLL int: PLL int = 0x0a;
							//b6 = by_pll: by_pll = 0;
							//b7 = EN_XCLK: chip-pin CLKI is pixel clock.
	buffer[24] = 0x0a;		//b0 = ni_out is always 0;
							//b1 = setup = 1 for NTSC;
							//b2 = 625line = 0 for NTSC;
							//b3 = vsync_dur = 1 for NTSC;
							//b4 = dic_screset is always 0;
							//b5 = pal_md = 0 for NTSC;
							//b6 = eclip is always 0;
							//b7 = reserved (en_scart) is always 0.
	buffer[25] = 0xe5;		//sync_amp $e5 for NTSC
	buffer[26] = 0x75;		//bst_amp $74-$76 for NTSC
	buffer[27] = 0x78;		//mcr: r-y $77-$79 for NTSC
	buffer[28] = 0x44;		//mcb: b-y $43-$44 for NTSC
	buffer[29] = 0x85;		//my: y $85 for NTSC
	buffer[30] = 0x37;		//lsb msc: msc b31-0: NTSC formula: ((3579545 / pixelclk) * 2^32) = MSC
	buffer[31] = 0x12;		//msc = $251b1237
	buffer[32] = 0x1b;
	buffer[33] = 0x25;		//msb msc.
	buffer[34] = 0x00;		//phase_off always $00

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 35); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return(err);
}//end InitBT_NTSC640_OS.

int ReadBT_ConnectStat (int fd, riva_data_t *ddata)
{
	int err, rd;
	
	int monstat;
	byte buffer[3];
	ddata->iic.buffer = &buffer;

	//set BT to return connection status in ESTATUS on next read CMD:
	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0xc4;		//set ESTATUS at b'01' (return conn.stat.);
	buffer[2] = 0x41;		//and leave chip outputs on.

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 3); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
		return(-1);
	}

	//do actual read connection status:
	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0xba;		//select register with CHECK_STAT CMD;
	buffer[2] = 0x40;		//give actual command.

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 3); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
		return(-1);
	}

	//CX: Wait 600uS for signals to stabilize (see datasheet)
	snooze(600);

	//read back updated connection status:
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit 1 byte
	buffer[0] = Adres + RD;	//issue IIC read command
	write(fd, ddata, 1); 
	//receive 1 byte
	//ACK level to TX after last byte to RX should be 1 (= NACK) (see IIC spec)
	//While the BT's don't care, CX chips will block the SDA line if an ACK gets sent!
	buffer[0] = 1;
	rd = read(fd, ddata, 1);
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (!rd==1 || err) //(not one byte RX'd) or (IIC comms error)?
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
		return(-1);
	}
	monstat = ((buffer[0] & 0xe0) >> 5);
	printf("\n(TV output monitor status = %d)\n",monstat);

	//instruct BT to go back to normal operation:
	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0xba;		//select register with CHECK_STAT CMD;
	buffer[2] = 0x00;		//give actual command.

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 3); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
		return (-1);
	}
	return (monstat);
}//end ReadBT_ConnectStat.

int KillClock_Blackout_BT (int fd, int type, riva_data_t *ddata)
{
	int err;
	byte buffer[4];

	ddata->iic.buffer = &buffer;

	if (type < 2) //BT...
	{
		//Only disable external pixelclock input on BT's.
		//CX chips will lock the bus if you do this.
		//(It looks like the external pixelclock is always OK as long as a valid 
		// mode is programmed for the TVout chip. This means that disabling the use
		// of this clock is not needed anyway.
		// If you do disable this input, this pixelclock will rise to about 60Mhz BTW..)

		//disable use of external pixelclock source...
		//(should prevent BT for being 'overclocked' by RIVA in VGA-only mode...)
		buffer[0] = Adres + WR;	//issue IIC write command
		buffer[1] = 0xa0;		//select BT register for setting EN_XCLK
		buffer[2] = 0x00;		//clear it
	
		//transmit it...
		//iicstart:
		ioctl(fd, rivaStartIIC, ddata);
		//transmit command string:
		write(fd, ddata, 3); 
		//iicstop:	
		err = ioctl(fd, rivaStopIIC, ddata);
		//errors encountered?
		if (err)
		{
			printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
			close (fd);
			return(err);
		}
	}
	else //CX...
	{
		//Disable CX video out (or wild output will be seen on TV..)
		buffer[0] = Adres + WR;	//issue IIC write command
		buffer[1] = 0x6c;		//select register in CX
		buffer[2] = 0x02;		//disable active video out.
	
		//transmit it...
		//iicstart:
		ioctl(fd, rivaStartIIC, ddata);
		//transmit command string:
		write(fd, ddata, 3); 
		//iicstop:	
		err = ioctl(fd, rivaStopIIC, ddata);
		//errors encountered?
		if (err)
		{
			printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
			close (fd);
			return(err);
		}
	}

	//black-out TVout while outputs are enabled...
	buffer[0] = Adres + WR;	//issue IIC write command
	buffer[1] = 0xc4;		//select first TV config register
	buffer[2] = 0x01;		//disable testimage while outputs remain enabled
	buffer[3] = 0x18;		//input is 24bit mpx'd RGB, BLANK = in, sync = act. hi

	//transmit it...
	//iicstart:
	ioctl(fd, rivaStartIIC, ddata);
	//transmit command string:
	write(fd, ddata, 4); 
	//iicstop:	
	err = ioctl(fd, rivaStopIIC, ddata);
	//errors encountered?
	if (err)
	{
		printf ("\n\nIIC communication error %d encountered, exiting...\n",	ddata->iic.status);
		close (fd);
	}
	return(err);
}//end KillClock_Blackout_BT.
