// VideoAddOn.cpp

#include <OS.h>
#include <File.h>
#include "VideoAddOn.h"

#include "Log.h"

const char * VideoAddOn::kSettingsFileName =
    "/boot/home/config/settings/BeTVOutVideoConsumer";

volatile port_id VideoAddOn::fVideoConsumerPort = 0;


VideoAddOn::VideoAddOn(
    image_id image)
 :  BMediaAddOn(image),
    fStatus(B_NO_INIT),
    fFlavors_count(0),
    fFormats_count(0),
    fConfigurations(),
    fConsumer(NULL)
{
	static char name[32] = { "BeTVOut Video Consumer" };
	static char info[64] = {
	    "BeTVOut Video Consumer, Copyright (C)2002 Kevin H. Patterson" };
	
	Log("VideoAddOn::VideoAddOn\n");
	
	/* Video Output parameters */
	fFlavors[0].name = name;
	fFlavors[0].info = info;
	fFlavors[0].kinds = B_BUFFER_CONSUMER | B_CONTROLLABLE | B_PHYSICAL_OUTPUT;
	fFlavors[0].flavor_flags = B_FLAVOR_IS_GLOBAL;
	fFlavors[0].internal_id = 7778;
	fFlavors[0].possible_count = 1;
	fFlavors[0].in_format_count = 10;
	fFlavors[0].in_format_flags = 0;
	fFlavors[0].in_formats = fFormats;
	fFlavors[0].out_format_count = 0;
	fFlavors[0].out_format_flags = 0;
	fFlavors[0].out_formats = NULL;
	
	fFlavors_count = 1;
	
	/* Acceptable Video Formats */
	fFormats[0].type = B_MEDIA_RAW_VIDEO;
	fFormats[0].u.raw_video = media_raw_video_format::wildcard;
	fFormats[0].u.raw_video.interlace = 1;
	fFormats[0].u.raw_video.display.format = B_RGB32;
	
	fFormats[1].type = B_MEDIA_RAW_VIDEO;
	fFormats[1].u.raw_video = media_raw_video_format::wildcard;
	fFormats[1].u.raw_video.interlace = 2;
	fFormats[1].u.raw_video.display.format = B_RGB32;
	
	fFormats[2].type = B_MEDIA_RAW_VIDEO;
	fFormats[2].u.raw_video = media_raw_video_format::wildcard;
	fFormats[2].u.raw_video.interlace = 1;
	fFormats[2].u.raw_video.display.format = B_RGB24;
	
	fFormats[3].type = B_MEDIA_RAW_VIDEO;
	fFormats[3].u.raw_video = media_raw_video_format::wildcard;
	fFormats[3].u.raw_video.interlace = 2;
	fFormats[3].u.raw_video.display.format = B_RGB24;
	
	fFormats[4].type = B_MEDIA_RAW_VIDEO;
	fFormats[4].u.raw_video = media_raw_video_format::wildcard;
	fFormats[4].u.raw_video.interlace = 1;
	fFormats[4].u.raw_video.display.format = B_RGB16;
	
	fFormats[5].type = B_MEDIA_RAW_VIDEO;
	fFormats[5].u.raw_video = media_raw_video_format::wildcard;
	fFormats[5].u.raw_video.interlace = 2;
	fFormats[5].u.raw_video.display.format = B_RGB16;
	
	fFormats[6].type = B_MEDIA_RAW_VIDEO;
	fFormats[6].u.raw_video = media_raw_video_format::wildcard;
	fFormats[6].u.raw_video.interlace = 1;
	fFormats[6].u.raw_video.display.format = B_RGB15;
	
	fFormats[7].type = B_MEDIA_RAW_VIDEO;
	fFormats[7].u.raw_video = media_raw_video_format::wildcard;
	fFormats[7].u.raw_video.interlace = 2;
	fFormats[7].u.raw_video.display.format = B_RGB15;
	
	fFormats[8].type = B_MEDIA_RAW_VIDEO;
	fFormats[8].u.raw_video = media_raw_video_format::wildcard;
	fFormats[8].u.raw_video.interlace = 1;
	fFormats[8].u.raw_video.display.format = B_GRAY8;
	
	fFormats[9].type = B_MEDIA_RAW_VIDEO;
	fFormats[9].u.raw_video = media_raw_video_format::wildcard;
	fFormats[9].u.raw_video.interlace = 2;
	fFormats[9].u.raw_video.display.format = B_GRAY8;
	
	fFormats_count = 10;
	
	//BETA: Some sort of check or other code should go here to see if the
	//video output hardware is available. fFlavors[].possible_count should
	//reflect the number of supported devices found.
	if (true) {
		LoadConfigurationFile();
		fStatus = B_OK;
	}
}


VideoAddOn::~VideoAddOn()
{
	Log("VideoAddOn::~VideoAddOn\n");
	
	if (fStatus == B_OK) {
		SaveConfigurationFile();
		//BETA: We may need to "uninitialize" the video output hardare here...
	}
}


status_t VideoAddOn::InitCheck( const char ** outFailureText )
{
	Log("VideoAddOn::InitCheck\n");
	
	if (fStatus != B_OK)
		*outFailureText = "Hardware not found!";
	
	return fStatus;
}


int32 VideoAddOn::CountFlavors()
{
	Log("VideoAddOn::CountFlavors\n");
	
	if (fStatus != B_OK)
		return fStatus;
	
	return fFlavors_count;
}


status_t VideoAddOn::GetFlavorAt( int32 n, const flavor_info ** outInfo )
{
	Log("VideoAddOn::GetFlavorAt\n");
	
	if (fStatus != B_OK)
		return fStatus;
	
	if (n >= 0 && n < fFlavors_count) {
		*outInfo = &fFlavors[n];
		return B_OK;
	}
	
	return B_BAD_INDEX;
}


BMediaNode * VideoAddOn::InstantiateNodeFor (
	const flavor_info * info,
	BMessage * /* configuration */,
	status_t * outError )
{
	Log("VideoAddOn::InstantiateNodeFor\n");
	
	if (fStatus != B_OK) {
		*outError = B_NO_INIT;
		return NULL;
	}
	
	if (info->internal_id == fFlavors[0].internal_id) {
		VideoConsumer * node = fConsumer;
		fConsumer = new VideoConsumer(
			this, fFlavors[0].name, fFlavors[0].internal_id);
			
		if (fConsumer != NULL && fConsumer->InitCheck() != B_OK) {
			*outError = fConsumer->InitCheck();
			delete fConsumer;
			fConsumer = node;
			return NULL;
		}
		
		#if 1
			//HACK: load configuration settings
			LoadConfigurationFor(fConsumer);
		#endif
		
		return fConsumer;
	}
	
	return NULL;
}


status_t VideoAddOn::LoadConfigurationFile()
{
	Log("VideoAddOn::LoadConfigurationFile\n");
	
	BFile file(kSettingsFileName, B_READ_ONLY);
	if (file.InitCheck() == B_OK)
		return fConfigurations.Unflatten(&file);
	
	return B_ERROR;
}


status_t VideoAddOn::SaveConfigurationFile()
{
	Log("VideoAddOn::SaveConfigurationFile\n");
	
	BFile file(kSettingsFileName, B_CREATE_FILE | B_WRITE_ONLY);
	if (file.InitCheck() == B_OK)
		return fConfigurations.Flatten(&file);
	
	return B_ERROR;
}


status_t VideoAddOn::LoadConfigurationFor( BMediaNode * node )
{
	Log("VideoAddOn::LoadConfigurationFor\n");
	
	int32 internal_id;
	if (node != NULL && node->AddOn(&internal_id) == this) {
		if (node == fConsumer && internal_id == fFlavors[0].internal_id) {
			BMessage message;
			if (fConfigurations.FindMessage("VideoConsumer", &message) == B_OK)
				fConsumer->SetConfiguration(&message);
			return B_OK;
		}
	}
	return B_ERROR;
}


status_t VideoAddOn::SaveConfigurationFor( const BMediaNode * node )
{
	Log("VideoAddOn::SaveConfigurationFor\n");
	
	int32 internal_id;
	if (node != NULL && node->AddOn(&internal_id) == this) {
		if (node == fConsumer && internal_id == fFlavors[0].internal_id) {
			BMessage message;
			fConsumer->GetConfiguration(&message);
			if (fConfigurations.HasMessage("VideoConsumer"))
				fConfigurations.ReplaceMessage("VideoConsumer", &message);
			else
				fConfigurations.AddMessage("VideoConsumer", &message);
			return B_OK;
	  }
	}
	return B_ERROR;
}


status_t VideoAddOn::GetConfigurationFor(
	BMediaNode * node,
	BMessage * configuration )
{
	Log("VideoAddOn::GetConfigurationFor\n");
	
	int32 internal_id;
	if (node->AddOn(&internal_id) == this) {
		if (node == fConsumer && internal_id == fFlavors[0].internal_id) {
			fConsumer->GetConfiguration(configuration);
			return B_OK;
		}
	}
	return B_ERROR;
}


bool VideoAddOn::WantsAutoStart()
{
	Log("VideoAddOn::WantsAutoStart\n");
	
	return false;
}


status_t VideoAddOn::AutoStart(
	int /* inCount */,
	BMediaNode ** /* outNode */,
	int32 * /*outInternalID */,
	bool * /* outHasMore */ )
{
	Log("VideoAddOn::AutoStart\n");
	
	return B_ERROR;
}


BMediaAddOn * make_media_addon( image_id image )
{
	return new VideoAddOn(image);
}
